#!/usr/bin/env python3
import struct

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.lines import Line2D
from matplotlib.patches import Circle

LINE_STYLES = ("solid", "dashed", "dotted", "3", "4", "5")

H_ALIGN = ("left", "center", "right")

V_ALIGN = (
    "<????>normal",
    "top",
    "<????>cap",
    "center",
    "baseline",
    "bottom",
)


class BinaryReader:
    def __init__(self, path):
        self.f = open(path, "rb")

    def readChar(self):
        return self.f.read(1)

    def readInt(self):
        return struct.unpack("i", self.f.read(4))[0]

    def readDouble(self):
        return struct.unpack("d", self.f.read(8))[0]

    def readBool(self):
        return struct.unpack("c", self.f.read(1))[0] == b"\x01"

    def readString(self, n=None):
        if n is None:
            n = self.readInt()
        return self.f.read(n).decode()

    def readDoubleArray(self, n):
        return np.fromfile(self.f, float, n)


class Layout:
    def __init__(self, x, y, width, height, minX, minY, maxX, maxY):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.minX = minX
        self.minY = minY
        self.maxX = maxX
        self.maxY = maxY

        # print('Layout', x, y, width, height, minX, minY, maxX, maxY)


class BinaryDecoder(BinaryReader):
    def __init__(self, path):
        super().__init__(path)
        self.current_colour = "k"
        self.current_linewidth = 1.0
        self.current_linestyle = "-"
        # self.layout = Layout(0.0, 0.0, 100.0, 100.0, 0.0, 0.0, 100.0, 100.0)
        self.stack = []

        #
        self.dimensionX = None  # Will be read from file
        self.dimensionY = None  # Will be read from file
        self.offsetX = 0.0
        self.offsetY = 0.0
        self.coordRatioX = 1.0
        self.coordRatioY = 1.0

    def text(self):
        # print("text")
        size = self.readInt()
        # assert size == 1
        r = self.readDouble()
        g = self.readDouble()
        b = self.readDouble()

        angle = -self.readDouble() * 180 / 3.1416
        blank = self.readBool()
        horizontal = self.readInt()
        vertical = self.readInt()

        n = self.readInt()

        texts = []

        for i in range(n):
            r = self.readDouble()
            g = self.readDouble()
            b = self.readDouble()
            s = self.readDouble()  # noqa
            m = self.readInt()
            texts.append(self.readString(m))

        for i in range(size):
            x = self.projectX(self.readDouble())
            y = self.projectY(self.readDouble())

            props = dict(
                ha=H_ALIGN[horizontal],
                va=V_ALIGN[vertical],
                rotation=angle,
                color=(r, g, b),
            )

            if blank:
                props["bbox"] = dict(
                    alpha=1,
                    facecolor="white",
                    #  pad=10,
                    edgecolor="none",
                )

            # print('text', x,y,texts[i])
            self.ax.text(
                x,
                y,
                texts[i],
                #    transform=ax.transAxes,
                #    fontsize=14,
                props,
            )

    def project(self):
        self.stack.append(
            (
                self.dimensionX,
                self.dimensionY,
                self.offsetX,
                self.offsetY,
                self.coordRatioX,
                self.coordRatioY,
            )
        )

        layout = Layout(
            self.readDouble(),
            self.readDouble(),
            self.readDouble(),
            self.readDouble(),
            self.readDouble(),
            self.readDouble(),
            self.readDouble(),
            self.readDouble(),
        )

        self.offsetX += layout.x * 0.01 * self.dimensionX
        self.offsetY += layout.y * 0.01 * self.dimensionY
        self.dimensionX = layout.width * 0.01 * self.dimensionX
        self.dimensionY = layout.height * 0.01 * self.dimensionY

        sumX = layout.maxX - layout.minX
        sumY = layout.maxY - layout.minY

        if sumX != 0 and sumY != 0:
            self.coordRatioX = self.dimensionX / sumX
            self.coordRatioY = self.dimensionY / sumY

        self.offsetX = self.projectX(-layout.minX)
        self.offsetY = self.projectY(-layout.minY)

    def projectX(self, x):
        return self.coordRatioX * x + self.offsetX

    def projectY(self, y):
        return self.coordRatioY * y + self.offsetY

    def unproject(self):
        (
            self.dimensionX,
            self.dimensionY,
            self.offsetX,
            self.offsetY,
            self.coordRatioX,
            self.coordRatioY,
        ) = self.stack.pop()
        # print("Pop")

    def colour(self):
        r = self.readDouble()
        g = self.readDouble()
        b = self.readDouble()
        a = self.readDouble()
        self.current_colour = (r, g, b, a)
        # print('colour', self.current_colour)

    def line_style(self):
        self.current_linestyle = LINE_STYLES[self.readInt()]
        self.current_linewidth = self.readDouble()

    def line_width(self):
        self.current_linewidth = self.readDouble()

    def new_page(self):
        pass

    def end_page(self):
        pass

    def arrows(self):
        assert False

    def flags(self):
        assert False

    def pixmap(self):
        assert False

    def image(self):
        assert False

    def poly_line(self):
        # print("poly_line")
        n = self.readInt()
        x = self.projectX(self.readDoubleArray(n))
        y = self.projectY(self.readDoubleArray(n))
        self.ax.add_line(
            Line2D(
                x,
                y,
                color=self.current_colour,
                linewidth=self.current_linewidth,
                linestyle=self.current_linestyle,
            )
        )
        self.ax.set_xlim(min(x), max(x))
        self.ax.set_ylim(min(y), max(y))

    def circle(self):

        x = self.projectX(self.readDouble())
        y = self.projectX(self.readDouble())
        r = self.readDouble()
        cs = self.readInt()  # noqa
        return
        self.ax.add_patch(
            Circle(
                (
                    x,
                    y,
                ),
                r,
                # color=self.current_colour,
                # linewidth=self.current_linewidth,
                # linestyle=self.current_linestyle,
            )
        )

    def simple_polygon(self):
        # print("simple_polygon")
        n = self.readInt()
        x = self.projectX(self.readDoubleArray(n))
        y = self.projectY(self.readDoubleArray(n))
        self.ax.add_line(
            Line2D(
                x,
                y,
                color=self.current_colour,
                linewidth=self.current_linewidth,
                linestyle=self.current_linestyle,
            )
        )

    def poly_line_2(self):
        # print("poly_line_2")
        n = self.readInt()
        x = self.projectX(self.readDoubleArray(n))
        y = self.projectY(self.readDoubleArray(n))
        self.ax.add_line(
            Line2D(
                x,
                y,
                color=self.current_colour,
                linewidth=self.current_linewidth,
                linestyle=self.current_linestyle,
            )
        )

    def simple_polygon_with_holes(self):
        # print("simple_polygon_with_holes")
        n = self.readInt()
        x = self.projectX(self.readDoubleArray(n))
        y = self.projectY(self.readDoubleArray(n))

        # Holes
        for j in range(self.readInt()):
            n = self.readInt()
            # hole_x =
            self.readDoubleArray(n)
            # hole_y =
            self.readDoubleArray(n)

        # Fill colour
        r = self.readDouble()
        g = self.readDouble()
        b = self.readDouble()
        a = self.readDouble()

        self.ax.fill(x, y, color=(r, g, b, a))

    def plot(self, axes):

        self.ax = axes

        DECODERS = {
            b"A": self.arrows,
            b"B": self.poly_line_2,
            b"C": self.colour,
            b"E": self.end_page,
            b"F": self.flags,
            b"H": self.poly_line,
            b"I": self.image,
            b"L": self.line_style,
            b"M": self.pixmap,
            b"N": self.new_page,
            b"P": self.project,
            b"R": self.circle,
            b"S": self.simple_polygon,
            b"T": self.text,
            b"U": self.unproject,
            b"W": self.line_width,
            b"X": self.simple_polygon_with_holes,
        }

        # Magic
        assert self.readString(6) == "MAGICS"

        # Endianess
        assert self.readInt() == 10

        # Version
        self.readInt()

        # header length
        self.readInt()

        self.dimensionX = self.readDouble()
        self.dimensionY = self.readDouble()

        print(self.dimensionX, self.dimensionY)

        op = self.readChar()
        while op:
            DECODERS[op]()
            op = self.readChar()

        # self.ax.set_xlim(min(x), max(x))
        # self.ax.set_ylim(min(y), max(y))


def plot_mgb(path):
    decoder = BinaryDecoder(path)
    decoder.plot(plt.gca())



# (C) Copyright 2012-2018 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

import ctypes
import ctypes.util
import json
import sys

import numpy as np
from numpy.ctypeslib import ndpointer

try:
    import ecmwflibs as findlibs
except ImportError:
    import findlibs

lib = findlibs.find("MagPlus")
if lib is None:
    raise RuntimeError("Cannot find MagPlus library")


dll = ctypes.CDLL(lib)


class FILE(ctypes.Structure):
    pass


FILE_p = ctypes.POINTER(FILE)


####################################################################


def get_library_path():
    return lib


####################################################################


def _string_to_char(x):
    return x.encode()


def _char_to_string(x):
    return x.decode()


def _convert_strings(fn):

    convert = False

    for a in fn.argtypes:
        if a is c_char_p:
            convert = True

    if fn.restype is c_char_p:
        convert = True

    if not convert:
        return fn

    def wrapped(*args):

        new_args = []
        for a, t in zip(args, fn.argtypes):
            if t is c_char_p:
                a = string_to_char(a)
            new_args.append(a)
        r = fn(*new_args)
        if fn.restype is c_char_p:
            r = char_to_string(r)
        return r

    return wrapped


def _noop(x):
    return x


if sys.version_info[0] > 2:
    convert_strings = _convert_strings
    char_to_string = _char_to_string
    string_to_char = _string_to_char
else:
    convert_strings = _noop
    char_to_string = _noop
    string_to_char = _noop


####################################################################
c_int = ctypes.c_int
c_int_p = ctypes.POINTER(c_int)

c_double = ctypes.c_double
c_double_p = ctypes.POINTER(c_double)

c_char = ctypes.c_char
c_char_p = ctypes.c_char_p

c_void_p = ctypes.c_void_p


####################################################################
def checked_error_in_last_paramater(fn):
    def wrapped(*args):
        err = c_int(0)
        err_p = ctypes.cast(ctypes.addressof(err), c_int_p)
        params = [a for a in args]
        params.append(err_p)

        result = fn(*params)
        if err.value:
            raise MagicsError(err)
        return result

    return wrapped


def checked_return_code(fn):
    def wrapped(*args):
        err = fn(*args)
        if err:
            raise MagicsError(char_to_string(err))

    return wrapped


####################################################################


def return_type(fn, ctype):
    def wrapped(*args):
        result = ctype()
        result_p = ctypes.cast(ctypes.addressof(result), ctypes.POINTER(ctype))
        params = [a for a in args]
        params.append(result_p)
        fn(*params)
        return result.value

    return wrapped


####################################################################

py_open = dll.py_open
py_open.restype = c_char_p


@checked_return_code
def init():
    return py_open()


####################################################################

py_close = dll.py_close
py_close.restype = c_char_p


@checked_return_code
def finalize():
    return py_close()


####################################################################

py_coast = dll.py_coast
py_coast.restype = c_char_p


@checked_return_code
def coast():
    return py_coast()


####################################################################

py_grib = dll.py_grib
py_grib.restype = c_char_p


@checked_return_code
def grib():
    return py_grib()


def oldversion():
    msg = "You are using an old version of magics ( < 4.0.0)"
    return msg.encode()


try:
    version = dll.version
    version.restype = ctypes.c_char_p
    version.argtypes = None
except Exception:
    version = oldversion

try:
    tile = dll.py_tile
except Exception:
    print("Tile not enabled: You are using an old version of magics ( < 4.1.0)")
    tile = oldversion


home = dll.home
home.restype = ctypes.c_char_p
home.argtypes = None

metanetcdf = dll.py_metanetcdf
metanetcdf.restype = ctypes.c_char_p
metanetcdf.argtypes = None

metagrib = dll.py_metagrib
metagrib.restype = ctypes.c_char_p
metagrib.argtypes = None

metainput = dll.py_metainput
metainput.restype = ctypes.c_char_p
metainput.argtypes = None


try:
    py_detect = dll.py_detect
    py_detect.restype = ctypes.c_char_p
    py_detect.argtypes = (ctypes.c_char_p, ctypes.c_char_p)
    py_detect = convert_strings(py_detect)

    detect = py_detect

except Exception:
    detect = dll.detect
    detect.restype = ctypes.c_char_p
    detect.argtypes = (ctypes.c_char_p, ctypes.c_char_p)
    detect = convert_strings(detect)

####################################################################

py_cont = dll.py_cont
py_cont.restype = c_char_p


@checked_return_code
def cont():
    return py_cont()


####################################################################
py_legend = dll.py_legend
py_legend.restype = c_char_p


@checked_return_code
def legend():
    return py_legend()


####################################################################

py_odb = dll.py_odb
py_odb.restype = c_char_p


@checked_return_code
def odb():
    return py_odb()


####################################################################

py_obs = dll.py_obs
py_obs.restype = c_char_p


@checked_return_code
def obs():
    return py_obs()


####################################################################


py_raw = dll.py_raw
py_raw.restype = c_char_p


@checked_return_code
def raw():
    return py_raw()


####################################################################

py_netcdf = dll.py_netcdf
py_netcdf.restype = c_char_p


@checked_return_code
def netcdf():
    return py_netcdf()


####################################################################

py_image = dll.py_image
py_image.restype = c_char_p


@checked_return_code
def image():
    return py_image()


####################################################################

py_plot = dll.py_plot
py_plot.restype = c_char_p


@checked_return_code
def plot():
    return py_plot()


####################################################################
py_text = dll.py_text
py_text.restype = c_char_p


@checked_return_code
def text():
    return py_text()


####################################################################

py_wind = dll.py_wind
py_wind.restype = c_char_p


@checked_return_code
def wind():
    return py_wind()


####################################################################

py_line = dll.py_line
py_line.restype = c_char_p


@checked_return_code
def line():
    return py_line()


####################################################################

py_symb = dll.py_symb
py_symb.restype = c_char_p


@checked_return_code
def symb():
    return py_symb()


####################################################################
py_boxplot = dll.py_boxplot
py_boxplot.restype = c_char_p


@checked_return_code
def boxplot():
    return py_boxplot()


####################################################################

py_taylor = dll.py_taylor
py_taylor.restype = c_char_p


@checked_return_code
def taylor():
    return py_taylor()


####################################################################

py_tephi = dll.py_tephi
py_tephi.restype = c_char_p


@checked_return_code
def tephi():
    return py_tephi()


####################################################################

py_graph = dll.py_graph
py_graph.restype = c_char_p


@checked_return_code
def graph():
    return py_graph()


####################################################################

py_axis = dll.py_axis
py_axis.restype = c_char_p


@checked_return_code
def axis():
    return py_axis()


####################################################################


py_geo = dll.py_geo
py_geo.restype = c_char_p


@checked_return_code
def geo():
    return py_geo()


####################################################################
py_import = dll.py_import
py_import.restype = c_char_p


@checked_return_code
def mimport():
    return py_import()


####################################################################
py_info = dll.py_info
py_info.restype = c_char_p


@checked_return_code
def info():
    return py_info()


####################################################################

py_input = dll.py_input
py_input.restype = c_char_p


@checked_return_code
def minput():
    return py_input()


####################################################################
py_eps = dll.py_eps
py_eps.restype = c_char_p


@checked_return_code
def eps():
    return py_eps()


####################################################################
#
# Please note: these two functions changed compared to the previous SWIG based Python interface
#

py_metgraph = dll.py_metgraph
py_metgraph.restype = c_char_p


@checked_return_code
def metgraph():
    return py_metgraph()


py_epsinput = dll.py_epsinput
py_epsinput.restype = c_char_p


@checked_return_code
def epsinput():
    return py_epsinput()


####################################################################
#
# Please note: this function was called mmetbufr to the previous SWIG based Python interface
#

py_metbufr = dll.py_metbufr
py_metbufr.restype = c_char_p


@checked_return_code
def metbufr():
    return py_metbufr()


####################################################################

py_epsgraph = dll.py_epsgraph
py_epsgraph.restype = c_char_p


@checked_return_code
def epsgraph():
    return py_epsgraph()


####################################################################

py_epscloud = dll.py_epscloud
py_epscloud.restype = c_char_p


@checked_return_code
def epscloud():
    return py_epscloud()


####################################################################
py_epslight = dll.py_epslight
py_epslight.restype = c_char_p


@checked_return_code
def epslight():
    return py_epslight()


####################################################################

py_epsplumes = dll.py_epsplumes
py_epsplumes.restype = c_char_p


@checked_return_code
def epsplumes():
    return py_epsplumes()


####################################################################
py_epswind = dll.py_epswind
py_epswind.restype = c_char_p


@checked_return_code
def epswind():
    return py_epswind()


####################################################################
py_epswave = dll.py_epswave
py_epswave.restype = c_char_p


@checked_return_code
def epswave():
    return py_epswave()


####################################################################

py_epsbar = dll.py_epsbar
py_epsbar.restype = c_char_p


@checked_return_code
def epsbar():
    return py_epsbar()


####################################################################

py_epsshading = dll.py_epsshading
py_epsshading.restype = c_char_p


@checked_return_code
def epsshading():
    return py_epsshading()


####################################################################

py_wrepjson = dll.py_wrepjson
py_wrepjson.restype = c_char_p


@checked_return_code
def wrepjson():
    return py_wrepjson()


####################################################################


py_geojson = dll.py_geojson
py_geojson.restype = c_char_p


@checked_return_code
def geojson():
    return py_geojson()


####################################################################

py_mapgen = dll.py_mapgen
py_mapgen.restype = c_char_p


@checked_return_code
def mapgen():
    return py_mapgen()


####################################################################


py_table = dll.py_table
py_table.restype = c_char_p


@checked_return_code
def mtable():
    return py_table()


####################################################################

py_seti = dll.py_seti
py_seti.restype = c_char_p
# py_seti.argtypes = (c_char,)


@checked_return_code
def seti(name, value):
    name = string_to_char(name)
    return py_seti(name, value)


def known_drivers():
    try:
        drivers = dll.py_knowndrivers()
        drivers = json.loads(drivers.decode())

        return drivers["drivers"]
    except Exception:
        return "known_drivers is not implemented in this version"


####################################################################
py_set1i = dll.py_set1i
py_set1i.restype = c_char_p
# py_set1i.argtypes = (c_char_p, c_int_p, c_int)


@checked_return_code
def set1i(name, data):
    #    array = np.empty((size,), dtype=np.float64)
    #    array_p = array.ctypes.data_as(c_double_p)
    #    _set1r(name, array_p, size)
    size = len(data)
    name = string_to_char(name)
    array_p = (ctypes.c_int * size)(*data)
    return py_set1i(ctypes.c_char_p(name), array_p, size)


####################################################################

array_2d_int = ndpointer(dtype=int, ndim=2, flags="CONTIGUOUS")
set2i = dll.py_set2i
set2i.restype = None
set2i.argtypes = (c_char_p, array_2d_int, c_int, c_int)
set2i = convert_strings(set2i)

####################################################################

setr = dll.py_setr
setr.restype = None
setr.argtypes = (c_char_p, c_double)
setr = convert_strings(setr)

####################################################################
py_set1r = dll.py_set1r
py_set1r.restype = c_char_p
# py_set1r.argtypes = (c_char_p, c_double_p, c_int)


@checked_return_code
def set1r(name, data):
    size = len(data)
    name = string_to_char(name)
    array_p = (ctypes.c_double * size)(*data)
    return py_set1r(ctypes.c_char_p(name), array_p, size)


####################################################################

array_2d_double = ndpointer(dtype=np.double, ndim=2, flags="CONTIGUOUS")
set2r = dll.py_set2r
set2r.restype = None
set2r.argtypes = (c_char_p, array_2d_double, c_int, c_int)
set2r = convert_strings(set2r)

####################################################################

setc = dll.py_setc
setc.restype = None
setc.argtypes = (c_char_p, c_char_p)
setc = convert_strings(setc)

####################################################################

py_set1c = dll.py_set1c
py_set1c.restype = c_char_p
# py_set1c.argtypes = (c_char_p, c_char_p, c_int)


@checked_return_code
def set1c(name, data):
    new_data = [string_to_char(s) for s in data]
    name = string_to_char(name)
    data_p = (c_char_p * (len(new_data)))(*new_data)
    return py_set1c(ctypes.c_char_p(name), data_p, len(new_data))


####################################################################

new_page = dll.py_new
new_page.restype = None
new_page.argtypes = (c_char_p,)
new_page = convert_strings(new_page)

####################################################################

reset = dll.py_reset
reset.restype = None
reset.argtypes = (c_char_p,)
reset = convert_strings(reset)

####################################################################


class MagicsError(Exception):
    def __init__(self, err):
        super(MagicsError, self).__init__(str(err))


####################################################################


def no_log(a, b):
    print(
        "Log listeners not handled in this version, consider using a version > 4.0.0 "
    )


def not_implemented():
    print("Not Implemented, consider upgrading a version > 4.4.0 ")


try:
    set_python = dll.py_set_python
    set_python.restype = None
    set_python.argtypes = None

    keep_compatibility = dll.py_keep_compatibility
    keep_compatibility.restype = None
    keep_compatibility.argtypes = None

    mute = dll.py_mute
    mute.restype = None
    mute.argtypes = None

    unmute = dll.py_unmute
    unmute.restype = None
    unmute.argtypes = None

    knowndrivers = dll.py_knowndrivers
    knowndrivers.restype = ctypes.c_char_p
    knowndrivers.argtypes = None

except Exception:
    set_python = not_implemented
    keep_compatibility = not_implemented
    mute = not_implemented
    unmute = not_implemented
    knowndrivers = not_implemented

try:
    strict_mode = dll.py_strict_mode
    strict_mode.restype = None
    strict_mode.argtypes = None
except Exception:
    strict_mode = not_implemented


log = ctypes.CFUNCTYPE(ctypes.c_void_p, ctypes.c_void_p, c_char_p)

try:
    warning_log = dll.mag_add_warning_listener
    warning_log.restype = None
    warning_log.argtypes = (ctypes.c_void_p, log)

    error_log = dll.mag_add_error_listener
    error_log.restype = None
    error_log.argtypes = (ctypes.c_void_p, log)

    info_log = dll.mag_add_info_listener
    info_log.restype = None
    info_log.argtypes = (ctypes.c_void_p, log)

    debug_log = dll.mag_add_debug_listener
    debug_log.restype = None
    debug_log.argtypes = (ctypes.c_void_p, log)
except Exception:
    error_log = no_log
    warning_log = no_log
    debug_log = no_log
    info_log = no_log


@log
def magics_log(data, msg):
    print("SUPER-->", msg.decode())
    return 0


# if __name__ == "__main__":
#    print "..."



#
# Copyright 2017-2018 European Centre for Medium-Range Weather Forecasts (ECMWF).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse


def selfcheck():
    from . import Magics

    print("Found: Magics %r." % Magics.version().decode())
    print("Library:", Magics.get_library_path())
    print("Magics home:", Magics.home().decode())
    print("Your system is ready.")


def main(argv=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("command", help="Supported commands: selfcheck.")
    args = parser.parse_args(args=argv)
    if args.command == "selfcheck":
        selfcheck()
    else:
        raise RuntimeError(
            "Command not recognised %r. See usage with --help." % args.command
        )


if __name__ == "__main__":
    main()



# (C) Copyright 1996-2016 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.

from . import macro


def substitute(default, user):
    out = default
    if user is not None:
        for key in user:
            out[key] = user[key]
    return out


def geoplot(
    data,
    contour=None,
    output=None,
    background=None,
    foreground=None,
    area=None,
    title=[],
):

    default = {
        "area": {},
        "contour": {},
        "background": {
            "map_coastline_sea_shade": "on",
            "map_coastline_sea_shade_colour": "rgb(81,81,81)",
            "map_coastline_land_shade": "on",
            "map_label": "off",
            "map_coastline_land_shade_colour": "rgb(113,113,113)",
        },
        "foreground": {
            "map_coastline_thickness": 2,
            "map_grid_line_style": "dash",
            "map_grid_colour": "rgb(143,166,183)",
            "map_label": "off",
            "map_coastline_colour": "rgb(143,166,183)",
        },
    }

    background = macro.mcoast(substitute(default["background"], background))
    foreground = macro.mcoast(substitute(default["foreground"], foreground))
    projection = macro.mmap(substitute(default["area"], area))
    contour = macro.mcont(substitute(default["contour"], contour))

    # Define the title
    title = macro.mtext(text_lines=title, text_font_size=0.8, text_justification="left")
    if output is None:
        return macro.plot(projection, background, data, contour, foreground, title)

    return macro.plot(output, projection, background, data, contour, foreground, title)


def xyplot(data, contour=None, output=None):

    default = {"contour": {}}

    # Setting the cartesian view
    projection = macro.mmap(
        subpage_map_projection="cartesian",
        subpage_x_automatic="on",
        subpage_y_automatic="on",
    )
    # Vertical axis
    vertical = macro.maxis(
        axis_orientation="vertical",
        axis_grid="on",
        axis_grid_colour="grey",
        axis_grid_thickness=1,
        axis_grid_line_style="dot",
    )

    # Horizontal axis
    horizontal = macro.maxis(
        axis_orientation="horizontal",
        axis_grid="on",
        axis_grid_colour="grey",
        axis_grid_thickness=1,
        axis_grid_line_style="dot",
    )

    # Define the graph
    contour = macro.mcont(substitute(default["contour"], contour))

    # Define the title
    title = macro.mtext(text_font_size=0.8, text_justification="left")
    if output is None:
        return macro.plot(
            output, projection, vertical, horizontal, data, contour, title
        )

    return macro.plot(output, projection, vertical, horizontal, data, contour, title)


def graph(x, y, title="", graph=None, colour="ecmwf_blue"):

    default = {
        "graph": {
            "graph_line_colour": "ecmwf_blue",
            "graph_line_thickness": 2,
        }
    }

    x[0] = x[0] * 1.0
    y[0] = y[0] * 1.0

    # Setting the cartesian view
    projection = macro.mmap(
        subpage_map_projection="cartesian",
        subpage_x_automatic="on",
        subpage_y_automatic="on",
    )
    # Vertical axis
    vertical = macro.maxis(
        axis_orientation="vertical",
        axis_grid="on",
        axis_grid_colour="grey",
        axis_grid_thickness=1,
        axis_grid_line_style="dot",
    )

    # Horizontal axis
    horizontal = macro.maxis(
        axis_orientation="horizontal",
        axis_grid="on",
        axis_grid_colour="grey",
        axis_grid_thickness=1,
        axis_grid_line_style="dot",
    )

    # define the input data
    input = macro.minput(input_x_values=x, input_y_values=y)

    # Define the graph
    graph = macro.mgraph(substitute(default["graph"], graph))
    # Define the title
    title = macro.mtext(
        text_lines=[title], text_font_size=0.8, text_justification="left"
    )
    return macro.plot(projection, vertical, horizontal, input, graph, title)


colour = "ecmwf_blue"
font_size = 0.35
defaults = {
    "epswind": {
        "projection": {
            "subpage_map_projection": "cartesian",
            "subpage_x_axis_type": "date",
            "subpage_x_automatic": "on",
            "subpage_y_axis_type": "regular",
            "subpage_y_automatic": "off",
            "subpage_y_max": 43200.0,
            "subpage_y_min": -43200.0,
        },
        "vertical_axis": {
            "axis_orientation": "vertical",
            "axis_grid": "on",
            "axis_grid_colour": "navy",
            "axis_grid_line_style": "dash",
            "axis_grid_reference_level": 0.0,
            "axis_grid_reference_thickness": 1,
            "axis_line": "on",
            "axis_line_colour": "navy",
            "axis_tick": "off",
            "axis_tick_label": "off",
        },
        "epswind": {
            "eps_rose_wind_border_colour": "Rgb(0.5000, 0.5000, 0.5000)",
            "eps_rose_wind_colour": "greenish_blue",
        },
    },
    "epswave": {
        "projection": {
            "subpage_map_projection": "cartesian",
            "subpage_x_axis_type": "date",
            "subpage_x_automatic": "on",
            "subpage_y_axis_type": "regular",
            "subpage_y_automatic": "off",
            "subpage_y_max": 43200.0,
            "subpage_y_min": -43200.0,
        },
        "vertical_axis": {
            "axis_orientation": "vertical",
            "axis_grid": "on",
            "axis_grid_colour": "navy",
            "axis_grid_line_style": "dash",
            "axis_grid_reference_level": 0.0,
            "axis_grid_reference_thickness": 1,
            "axis_line": "on",
            "axis_line_colour": "navy",
            "axis_tick": "off",
            "axis_tick_label": "off",
        },
        "epswave": {
            "eps_rose_wind_border_colour": "Rgb(0.5000, 0.5000, 0.5000)",
            "eps_rose_wind_colour": "RGB(0.925,0.609,0.953)",
            "eps_rose_wind_convention": "oceanographic",
        },
    },
    "eps": {
        "projection": {
            "subpage_map_projection": "cartesian",
            "subpage_x_axis_type": "date",
            "subpage_x_automatic": "on",
            "subpage_y_axis_type": "regular",
            "subpage_y_automatic": "on",
        },
        "vertical_axis": {
            "axis_orientation": "vertical",
            "axis_grid": "on",
            "axis_grid_colour": "navy",
            "axis_grid_line_style": "dash",
            "axis_grid_reference_level": 0.0,
            "axis_grid_reference_thickness": 1,
            "axis_line": "on",
            "axis_line_colour": "navy",
            "axis_tick_colour": "navy",
            "axis_tick_label_colour": "navy",
            "axis_tick_label_height": font_size,
        },
        "horizontal_axis": {
            "axis_orientation": "horizontal",
            "axis_date_type": "days",
            "axis_days_label": "both",
            "axis_days_label_colour": "navy",
            "axis_days_label_height": font_size,
            "axis_grid": "on",
            "axis_grid_colour": "navy",
            "axis_grid_line_style": "dash",
            "axis_line_colour": "navy",
            "axis_minor_tick": "off",
            "axis_minor_tick_colour": "navy",
            "axis_months_label": "off",
            "axis_tick_colour": "navy",
            "axis_type": "date",
            "axis_years_label": "off",
        },
        "epsgraph": {
            "eps_box_border_thickness": 2,
            "eps_box_width": 1.5,
            "eps_box_colour": colour,
            "eps_font_colour": "navy",
            "eps_legend_font_size": font_size,
            "eps_grey_legend": "off",
            "legend": "off",
        },
        "epsclim": {
            "eps_shade_colour": colour,
            "eps_shade_line_thickness": 4,
        },
    },
}


def epswave(parameter, input, **args):
    actions = []

    projection = macro.mmap(
        substitute(defaults["epswave"]["projection"], args.get("projection", None))
    )

    # define horizontal axis
    horizontal = macro.maxis(
        substitute(
            defaults["eps"]["horizontal_axis"], args.get("horizontal_axis", None)
        )
    )
    vertical = macro.maxis(
        substitute(
            defaults["epswave"]["vertical_axis"], args.get("vertical_axis", None)
        )
    )

    data = macro.mwrepjson(
        wrepjson_family="eps",
        wrepjson_keyword="eps",
        wrepjson_input_filename=input,
        wrepjson_parameter=parameter,
        wrepjson_parameter_information=args.get("title", parameter),
        wrepjson_parameter_scaling_factor=1.0,
    )

    wave = macro.mepswave(
        substitute(defaults["epswave"]["epswave"], args.get("epswave", None))
    )

    actions.append(projection)
    actions.append(vertical)
    actions.append(horizontal)

    actions.append(data)
    actions.append(wave)

    text = macro.mtext(
        text_colour="navy",
        text_font_size=font_size * 2,
        text_justification="left",
        text_lines=[
            "ENS Meteogram",
            "<json_info key='station_name'/><json_info key='location'/>",
            "<json_info key='grid_point'/><json_info key='height'/>",
            "<json_info key='product_info'/><json_info key='date'/>",
            "<font size='0.5' colour='white'>.</font>",
            "<json_info key='parameter_info'/>",
        ],
    )

    actions.append(text)

    if "output" in args != "":
        # Setting of the output file name
        png = macro.output(
            output_formats=["png"],
            output_name_first_page_number="off",
            output_name=args["output"],
            super_page_y_length=10.0,
            subpage_y_length=5.0,
            subpage_y_position=1.5,
        )

        return macro._plot(png, actions)

    return macro.plot(actions)


def epswind(parameter, input, **args):
    actions = []

    projection = macro.mmap(
        substitute(defaults["epswind"]["projection"], args.get("projection", None))
    )

    # define horizontal axis
    horizontal = macro.maxis(
        substitute(
            defaults["eps"]["horizontal_axis"], args.get("horizontal_axis", None)
        )
    )
    vertical = macro.maxis(
        substitute(
            defaults["epswind"]["vertical_axis"], args.get("vertical_axis", None)
        )
    )

    data = macro.mwrepjson(
        wrepjson_family="eps",
        wrepjson_keyword="eps",
        wrepjson_input_filename=input,
        wrepjson_parameter=parameter,
        wrepjson_parameter_information=args.get("title", parameter),
        wrepjson_parameter_scaling_factor=1.0,
    )

    wave = macro.mepswind(
        substitute(defaults["epswind"]["epswind"], args.get("epswind", None))
    )

    actions.append(projection)
    actions.append(vertical)
    actions.append(horizontal)

    actions.append(data)
    actions.append(wave)

    text = macro.mtext(
        text_colour="navy",
        text_font_size=font_size * 2,
        text_justification="left",
        text_lines=[
            "ENS Meteogram",
            "<json_info key='station_name'/><json_info key='location'/>",
            "<json_info key='grid_point'/><json_info key='height'/>",
            "<json_info key='product_info'/><json_info key='date'/>",
            "<font size='0.5' colour='white'>.</font>",
            "<json_info key='parameter_info'/>",
        ],
    )

    actions.append(text)

    if "output" in args != "":
        # Setting of the output file name
        png = macro.output(
            output_formats=["png"],
            output_name_first_page_number="off",
            output_name=args["output"],
            super_page_y_length=10.0,
            subpage_y_length=5.0,
            subpage_y_position=1.5,
        )

        return macro._plot(png, actions)

    return macro.plot(actions)


def epsgraph(parameter, input, **args):

    actions = []

    projection = macro.mmap(
        substitute(defaults["eps"]["projection"], args.get("projection", None))
    )

    # define horizontal axis
    horizontal = macro.maxis(
        substitute(
            defaults["eps"]["horizontal_axis"], args.get("horizontal_axis", None)
        )
    )
    vertical = macro.maxis(
        substitute(defaults["eps"]["vertical_axis"], args.get("vertical_axis", None))
    )

    data = macro.mwrepjson(
        wrepjson_family="eps",
        wrepjson_keyword="eps",
        wrepjson_input_filename=input,
        wrepjson_parameter=parameter,
        wrepjson_missing_value=args.get("missing", 9999.0),
        wrepjson_parameter_information=args.get("title", parameter),
        wrepjson_parameter_scaling_factor=args.get("scaling", 1.0),
        wrepjson_parameter_offset_factor=args.get("offset", 0.0),
    )

    graph = macro.mepsgraph(
        substitute(defaults["eps"]["epsgraph"], args.get("epsgraph", None))
    )
    actions.append(projection)
    actions.append(vertical)
    actions.append(horizontal)

    if "climate" in args:
        clim = macro.mwrepjson(
            wrepjson_family="eps",
            wrepjson_keyword="clim",
            wrepjson_input_filename=input,
            wrepjson_parameter=parameter,
            wrepjson_parameter_scaling_factor=args.get("scaling", 1.0),
            wrepjson_parameter_offset_factor=args.get("offset", 0.0),
            wrepjson_ignore_keys=["100"],
            wrepjson_missing_value=args.get("missing", 9999.0),
            wrepjson_parameter_information="none",
            wrepjson_position_information="off",
        )
        shade = macro.mepsshading(
            substitute(defaults["eps"]["epsclim"], args.get("epsclim", None))
        )
        actions.append(clim)
        actions.append(shade)

    actions.append(data)
    actions.append(graph)

    text = macro.mtext(
        text_colour="navy",
        text_font_size=font_size * 2,
        text_justification="left",
        text_lines=[
            "ENS Meteogram",
            "<json_info key='station_name'/><json_info key='location'/>",
            "<json_info key='grid_point'/><json_info key='height'/>",
            "<json_info key='product_info'/><json_info key='date'/>",
            "<font size='0.5' colour='white'>.</font>",
            "<json_info key='parameter_info'/>",
        ],
    )

    actions.append(text)

    if "output" in args != "":
        # Setting of the output file name
        png = macro.output(
            output_formats=["png"],
            output_name_first_page_number="off",
            output_name=args["output"],
            super_page_y_length=10.0,
            subpage_y_length=5.0,
            subpage_y_position=1.0,
        )

        return macro._plot(png, actions)

    return macro.plot(actions)


def epsclimgram(**kw):

    args = {"clim": True}
    args.update(kw)

    return epsgraph(**args)


params = {
    "2t": {
        "scaling": 1.0,
        "offset": -273.0,
        "method": epsgraph,
        "title": "2-metre temperature",
    },
    "mcc": {
        "scaling": 1.0,
        "offset": 0.0,
        "method": epsgraph,
        "title": "Medium cloud cover",
    },
    "light-index": {
        "scaling": 1.0,
        "offset": 0.0,
        "method": epsgraph,
        "title": "Light index",
    },
    "10fg6": {
        "scaling": 1.0,
        "offset": 0.0,
        "method": epsgraph,
        "title": "10 metre wind gust in the last 6 hours",
    },
    "tp": {
        "scaling": 1000.0,
        "offset": 0.0,
        "method": epsgraph,
        "title": "Total precipitation (last 6 hours) (m)",
    },
    "tcc": {
        "scaling": 1.0,
        "offset": 0.0,
        "method": epsgraph,
        "title": "Total cloud cover",
    },
    "hcc": {
        "scaling": 1.0,
        "offset": 0.0,
        "method": epsgraph,
        "title": "High cloud cover",
    },
    "ws": {
        "scaling": 1.0,
        "offset": 0.0,
        "method": epsgraph,
        "title": "Wind speed (m/s)",
    },
    "lcc": {
        "scaling": 1.0,
        "offset": 0.0,
        "method": epsgraph,
        "title": "Low cloud cover",
    },
    "sf": {
        "scaling": 1000.0,
        "offset": 0.0,
        "method": epsgraph,
        "title": "Snowfall (last 6 hours)",
    },
    "swh": {
        "scaling": 1.0,
        "offset": 0.0,
        "method": epsgraph,
        "title": "Significant wave height",
    },
    "mwp": {
        "scaling": 1.0,
        "offset": 0.0,
        "method": epsgraph,
        "title": "Mean wave period",
    },
    "dwi": {
        "scaling": 1.0,
        "offset": 0.0,
        "method": epswind,
        "title": "10 metre wind direction",
    },
    "mwd": {
        "scaling": 1.0,
        "offset": 0.0,
        "method": epswave,
        "title": "Mean wave direction",
    },
    "ws24": {
        "scaling": 1.0,
        "offset": 0.0,
        "method": epsgraph,
        "title": "Daily mean of 10m wind speed (m/s)",
    },
    "tcc24": {
        "scaling": 1.0,
        "offset": 0.0,
        "method": epsgraph,
        "title": "Daily mean of total cloud cover",
    },
    "mn2t24": {
        "scaling": 1.0,
        "offset": -273.15,
        "method": epsgraph,
        "title": "2 metre min temperature (Daily)",
    },
    "tp24": {
        "scaling": 1000.0,
        "offset": 0.0,
        "method": epsgraph,
        "title": "Total precipitation (mm/24hr)",
    },
    "mx2t24": {
        "scaling": 1.0,
        "offset": -273.15,
        "method": epsgraph,
        "title": "2 metre max. temperature (Daily)",
    },
    "dwi24": {
        "scaling": 1.0,
        "offset": 0.0,
        "method": epswind,
        "title": "Daily distribution of 10m Wind Direction",
    },
}


params_15days = {
    "mx2t24": "2 metre max. temperature (Daily)",
    "mn2t24": "2 metre min temperature (Daily)",
    "tcc24": "Daily mean of total cloud cover",
    "tp24": "Total precipitation (mm/24hr)",
    "dwi24": "Daily distribution of 10m Wind Direction",
    "ws24": "Daily mean of 10m wind speed (m/s)",
}


def epsgram(parameter, input, **args):
    eps = params.get(
        parameter,
        {"scaling": 1.0, "offset": 0.0, "method": epsgraph, "title": parameter},
    )

    args["scaling"] = eps["scaling"]
    args["offset"] = eps["offset"]
    args["title"] = eps["title"]

    return eps["method"](parameter, input, **args)



# (C) Copyright 1996-2016 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.

import json
import os
import sys
import tempfile
import threading

import numpy

from . import Magics

LOCK = threading.RLock()

ipython_active = None
try:
    from IPython import get_ipython

    ipython_active = get_ipython()
except Exception:
    pass


class Context(object):
    def __init__(self):
        with LOCK:
            self.tmp = []
            Magics.set_python()
            self.silent = True

    def set(self):
        with LOCK:
            if self.silent:
                os.environ["MAGPLUS_WARNING"] = "off"
                Magics.mute()
            else:
                os.environ["MAGPLUS_INFO"] = "on"


global context

context = Context()


def silent():
    context.silent = False


def unmute():
    context.silent = False


def mute():
    context.silent = True


def keep_compatibility():
    with LOCK:
        Magics.keep_compatibility()


@Magics.log
def warning(int, msg):
    try:
        print(msg.decode().rstrip())
    except Exception:
        print(msg)


@Magics.log
def error(int, msg):
    try:
        print(msg.decode().rstrip())
    except Exception:
        print(msg)


@Magics.log
def info_log(int, msg):
    try:
        print(msg.decode().rstrip())
    except Exception:
        print(msg)


@Magics.log
def debug_log(int, msg):
    try:
        print(msg.decode().rstrip())
    except Exception:
        print(msg)


Magics.warning_log(3, warning)
Magics.error_log(3, error)


def debug():
    Magics.debug_log(3, debug_log)


def info():
    Magics.info_log(3, info_log)


actions = {
    "mobs": "pobs",
    "mcoast": "pcoast",
    "pcoast": "pcoast",
    "mtext": "ptext",
    "ptext": "ptext",
    "psymb": "psymb",
    "msymb": "psymb",
    "pcont": "pcont",
    "mcont": "pcont",
    "pgeo": "pgeo",
    "mgeojson": "pgeojson",
    "mgeo": "pgeo",
    "mlegend": "",
    "plegend": "",
    "mgrib": "pgrib",
    "pgrib": "pgrib",
    "mwind": "pwind",
    "pwind": "pwind",
    "mgraph": "pgraph",
    "pgraph": "pgraph",
    "maxis": "paxis",
    "paxis": "paxis",
    "minput": "pinput",
    "mtable": "ptable",
    "ptable": "ptable",
    "pboxplot": "pboxplot",
    "mboxplot": "pboxplot",
    "pinput": "pinput",
}


class Action(object):
    def __init__(self, verb, action, html, args):
        self.verb = verb
        self.action = action

        if verb == "output" and not ("output_formats" in args):
            args["output_formats"] = ["png"]

        self.args = args
        if html == "":
            self.html = verb
        else:
            self.html = "<a href=/wiki/display/MAGP/%s target='_blank'>%s</a>" % (
                html,
                verb,
            )

    def __repr__(self):
        x = ""
        for key in list(self.args.keys()):
            x = x + " %s = '%s'\n" % (key, self.args[key])
        return x

    def inspect(self):
        print(self)

    def quote(self, v):
        return '"' + v + '"'

    def to_yaml(self):
        def tidy(x):
            if isinstance(x, (list, tuple)):
                return [tidy(y) for y in x]

            if isinstance(x, dict):
                d = {}
                for k, v in x.items():
                    d[k] = tidy(v)
                return d

            # Numpy arrays
            if hasattr(x, "tolist"):
                return x.tolist()

            if x in ("on", "true", "yes"):
                return True

            if x in ("off", "false", "no"):
                return False

            return x

        return {self.verb: tidy(self.args)}

    def clean_object(self, obj):
        if sys.version_info[0] < 3:
            if type(obj) in (int, float, str, bool, numpy.float64, numpy.float32):
                return obj
            elif type(obj) == unicode:  # noqa
                return str(obj)
            elif type(obj) in (list, tuple, set, numpy.ndarray) and len(obj):
                if type(obj[0]) != unicode:  # noqa
                    return obj
                obj = list(obj)
                for i, v in enumerate(obj):
                    obj[i] = self.clean_object(v)
            elif type(obj) == dict:
                for i, v in list(obj.items()):
                    obj[i] = self.clean_object(v)
            else:
                print("Invalid object in data, converting to string: ")
                print(type(obj))
                obj = str(obj)
        return obj

    def find_type(self, data):
        for v in data:
            if not isinstance(v, int):
                return "float"
        return "int"

    def set(self):  # noqa C901
        for key in list(self.args.keys()):

            if isinstance(self.args[key], dict):
                Magics.setc(key, json.dumps(self.args[key]))
            elif isinstance(self.args[key], bool):
                if self.args[key]:
                    Magics.setc(key, "on")
                else:
                    Magics.setc(key, "off")
            elif isinstance(self.args[key], str):
                if key == "odb_data":
                    Magics.setc("odb_filename", self.args[key])
                else:
                    Magics.setc(key, self.args[key])
            elif isinstance(self.args[key], int):
                Magics.seti(key, self.args[key])
            elif isinstance(self.args[key], float):
                Magics.setr(key, self.args[key])
            elif isinstance(self.args[key], list) and len(self.args[key]):
                if isinstance(self.args[key][0], str):
                    Magics.set1c(key, self.args[key])
                elif isinstance(self.args[key][0], dict):
                    np = []
                    for p in self.args[key]:
                        np.append(json.dumps(p))
                    Magics.set1c(key, np)
                else:
                    type = self.find_type(self.args[key])
                    if type == "int":
                        Magics.set1i(key, numpy.array(self.args[key], dtype="i"))
                    else:
                        Magics.set1r(key, numpy.array(self.args[key]))
            elif isinstance(self.args[key], numpy.ndarray):
                type = self.args[key].dtype
                data = self.args[key].copy()
                size = data.shape
                dim = len(size)
                type = self.find_type(self.args[key])
                if type == "int":
                    if dim == 2:
                        Magics.set2i(key, data.astype(numpy.int64), size[0], size[1])
                    else:
                        Magics.set1i(key, data.astype(numpy.int64), size[0])
                elif type == "float":
                    if dim == 2:
                        Magics.set2r(key, data.astype(numpy.float64), size[1], size[0])
                    else:
                        Magics.set1r(key, data.astype(numpy.float64))
                else:
                    print("can not interpret type %s for %s ???->", (type, key))
            else:
                self.args[key].execute(key)

    def execute(self):

        if self.action != Magics.odb:
            self.args = self.clean_object(self.args)

        self.set()

        if self.action is not None:
            if self.action != Magics.new_page:
                if self.action == Magics.legend:
                    Magics.setc("legend", "on")
                self.action()
                if self.action != Magics.obs and self.action != Magics.minput:
                    for key in list(self.args.keys()):
                        Magics.reset(key)
            else:
                self.action("page")

    def style(self):

        if self.action not in [Magics.grib, Magics.netcdf, Magics.minput]:
            return {}

        self.args = self.clean_object(self.args)

        self.set()
        if self.action == Magics.grib:
            return Magics.metagrib()
        if self.action == Magics.netcdf:
            return Magics.metanetcdf()
        if self.action == Magics.minput:
            return Magics.metainput()


def encode_numpy(np_obj):
    """
    Encode numpy objects to their python equivalents.
    e.g. numpy.int32 -> int

    This is necessary because json is unable to serialize numpy types natively.
    """
    if isinstance(np_obj, numpy.ndarray):
        # Nested array elements are cast from numpy.generic to Python object
        return np_obj.tolist()
    elif isinstance(np_obj, numpy.generic):
        return np_obj.item()
    else:
        raise TypeError(
            "Object of type '{}' is not JSON serializable".format(type(np_obj))
        )


def detect(attributes, dimension):
    return Magics.detect(json.dumps(attributes, default=encode_numpy), dimension)


def detect_lat_lon(xarray_dataset, ds_attributes):
    attrs = {
        ds_attribute: xarray_dataset[ds_attribute].attrs
        for ds_attribute in ds_attributes
    }
    lat_name = detect(attrs, "latitude")
    lon_name = detect(attrs, "longitude")
    return lat_name, lon_name


def mxarray(xarray_dataset, xarray_variable_name, xarray_dimension_settings={}):
    """
    Convert an xarray dataset containing a variable with latitude and longitude data into
    magics.minput.
    """
    # usually we find latitude and longitude in xarray_dataset.coords, but we sometimes see 2d
    # lat/lon data in xarray_dataset.data_vars instead.
    for ds_attributes in [xarray_dataset.coords, xarray_dataset.data_vars]:
        ret = _mxarray(
            xarray_dataset,
            xarray_variable_name,
            ds_attributes,
            xarray_dimension_settings,
        )
        if ret:
            return ret

    raise ValueError("Could not find latitude and longitude in dataset")


def _mxarray(
    xarray_dataset, xarray_variable_name, ds_attributes, xarray_dimension_settings
):
    lat_name, lon_name = detect_lat_lon(xarray_dataset, ds_attributes)

    if lat_name and lon_name:
        lat_dim_names = sorted(xarray_dataset[lat_name].dims)
        lon_dim_names = sorted(xarray_dataset[lon_name].dims)
        n_lat_dims = len(lat_dim_names)
        n_lon_dims = len(lon_dim_names)

        if n_lat_dims != n_lon_dims:
            raise ValueError(
                "Dimension mismatch for latitude and longitude. "
                "lat_dim_names={} lon_dim_names={}".format(lat_dim_names, lon_dim_names)
            )
        elif n_lat_dims == 1:
            return _mxarray_1d(
                xarray_dataset,
                xarray_variable_name,
                lat_name,
                lon_name,
                xarray_dimension_settings,
            )
        elif n_lat_dims == 2:
            return _mxarray_2d(
                xarray_dataset,
                xarray_variable_name,
                lat_name,
                lon_name,
                xarray_dimension_settings,
                lat_dim_names,
            )
        else:
            raise ValueError(
                "Found latitude and longitude with more than 2 dimensions. "
                "lat_dim_names={} lon_dim_names={}".format(lat_dim_names, lon_dim_names)
            )


def _mxarray_1d(
    xarray_dataset, xarray_variable_name, lat_name, lon_name, xarray_dimension_settings
):
    lat = xarray_dataset[lat_name].values.astype(numpy.float64)
    lon = xarray_dataset[lon_name].values.astype(numpy.float64)
    input_field_values = _mxarray_flatten(
        xarray_dataset[xarray_variable_name],
        xarray_dimension_settings,
        [lat_name, lon_name],
    ).values.astype(numpy.float64)

    data = minput(
        input_field=input_field_values,
        input_latitudes_list=lat,
        input_longitudes_list=lon,
        input_metadata=dict(xarray_dataset[xarray_variable_name].attrs),
    )
    return data


def _mxarray_2d(
    xarray_dataset,
    xarray_variable_name,
    lat_name,
    lon_name,
    xarray_dimension_settings,
    dims_to_ignore,
):
    lat = xarray_dataset[lat_name].values.astype(numpy.float64)
    lon = xarray_dataset[lon_name].values.astype(numpy.float64)
    input_field_values = _mxarray_flatten(
        xarray_dataset[xarray_variable_name], xarray_dimension_settings, dims_to_ignore
    ).values.astype(numpy.float64)

    data = minput(
        input_field=input_field_values,
        input_field_organization="nonregular",
        input_field_latitudes=lat,
        input_field_longitudes=lon,
        input_metadata=dict(xarray_dataset[xarray_variable_name].attrs),
    )
    return data


def _mxarray_flatten(xarray_dataset, dims_to_flatten, dims_to_ignore):
    # flatten an nD matrix into a 2d matrix by slicing the matrix based on the values given to
    # dimensions in dims_to_flatten.
    for dim in xarray_dataset.dims:
        if dim in dims_to_ignore:
            continue
        elif dim in dims_to_flatten:
            if dims_to_flatten[dim] not in xarray_dataset[dim]:
                raise ValueError(
                    "Dimension not valid. dimension={} dtype={} options={} dtype={}".format(
                        dim,
                        type(dim),
                        xarray_dataset[dim].values,
                        xarray_dataset[dim].dtype,
                    )
                )
            else:
                xarray_dataset = xarray_dataset.loc[{dim: dims_to_flatten[dim]}]
        elif xarray_dataset[dim].size == 1:
            # automatically squash this dimension
            d = xarray_dataset[dim].values[0]
            print("automatically squashing dimension: {}={}".format(dim, d))
            xarray_dataset = xarray_dataset.loc[{dim: d}]
        else:
            raise ValueError(
                "Missing dimension to flatten. "
                "Please pick a dimension from which to slice data. "
                "dimension={} options={} dtype={}".format(
                    dim, xarray_dataset[dim].values, xarray_dataset[dim].dtype
                )
            )
    return xarray_dataset


def make_action(verb, action, html=""):
    def f(_m=None, **kw):
        args = {}
        if _m is not None:
            args.update(_m)
        args.update(kw)
        return Action(verb, action, html, args)

    return f


mcoast = make_action("mcoast", Magics.coast, "Coastlines")
pcoast = make_action("pcoast", Magics.coast)
maxis = make_action("maxis", Magics.axis, "Axis")
paxis = make_action("paxis", Magics.axis)
mcont = make_action("mcont", Magics.cont, "Contouring")
pcont = make_action("pcont", Magics.cont)
msymb = make_action("msymb", Magics.symb, "Symbol")
psymb = make_action("psymb", Magics.symb)
pimport = make_action("pimport", Magics.mimport)
mimport = make_action("mimport", Magics.mimport)
mtaylor = make_action("mtaylor", Magics.taylor)
mgeo = make_action("mgeo", Magics.geo)
pgeo = make_action("pgeo", Magics.geo)
pgrib = make_action("pgrib", Magics.grib, "Grib+Input")
mgrib = make_action("mgrib", Magics.grib, "Grib+Input")
pmapgen = make_action("pmapgen", Magics.mapgen)
mmapgen = make_action("mmapgen", Magics.mapgen)
pnetcdf = make_action("pnetcdf", Magics.netcdf)
mnetcdf = make_action("mnetcdf", Magics.netcdf)
odb_geopoints = make_action("odb_geopoints", Magics.odb, "Odbviewer")
odb_geovectors = make_action("odb_geovectors", Magics.odb, "Odbviewer")
odb_xypoints = make_action("odb_xypoints", Magics.odb, "Odbviewer")
odb_xyvectors = make_action("odb_xyvectors", Magics.odb, "Odbviewer")
pmap = make_action("pmap", None, "Subpage")
mmap = make_action("mmap", None, "Subpage")
plegend = make_action("plegend", Magics.legend, "Legend")
mlegend = make_action("mlegend", Magics.legend, "Legend")
ptext = make_action("mtext", Magics.text)
mtext = make_action("mtext", Magics.text)
output = make_action("output", None, "PNG Output")
pwind = make_action("pwind", Magics.wind, "Wind+Plotting")
mwind = make_action("mwind", Magics.wind, "Wind+Plotting")
pline = make_action("pline", Magics.line)
mline = make_action("mline", Magics.line)
pgraph = make_action("pgraph", Magics.graph, "Graph+Plotting")
mgraph = make_action("mgraph", Magics.graph, "Graph+Plotting")
pboxplot = make_action("pboxplot", Magics.boxplot)
mboxplot = make_action("mboxplot", Magics.boxplot)
pobs = make_action("pobs", Magics.obs)
mobs = make_action("mobs", Magics.obs)
page = make_action("page", Magics.new_page)
pinput = make_action("pinput", Magics.minput)
minput = make_action("minput", Magics.minput, "Input+Data")
mtable = make_action("mtable", Magics.mtable, "CSV+Table+Decoder")
mgeojson = make_action("mgeojson", Magics.geojson, "GeoJSon")
mwrepjson = make_action("mwrepjson", Magics.wrepjson, "WrepJSon")
mepsinput = make_action("mepsinput", Magics.epsinput, "EpsInput")
mepscloud = make_action("mepscloud", Magics.epscloud)
mepslight = make_action("mepslight", Magics.epslight)
mepsbar = make_action("mepsbar", Magics.epsbar)
mepswind = make_action("mepswind", Magics.epswind)
mepswave = make_action("mepswave", Magics.epswave)
mepsshading = make_action("mepsshading", Magics.epsshading)
mepsgraph = make_action("mepsgraph", Magics.epsgraph)
mepsplumes = make_action("mepsplumes", Magics.epsplumes)
mtephi = make_action("mtephi", Magics.tephi)
mtile = make_action("mtile", Magics.tile)

mmetgraph = make_action("mmetgraph", Magics.metgraph)
mmetbufr = make_action("mmetbufr", Magics.metbufr)


def examine(*args):
    for n in args:
        try:
            n.inspect()
        except Exception:
            break


def _execute(o):
    if isinstance(o, (list, tuple)):
        for x in o:
            _execute(x)
    else:
        o.execute()


def _plot(*args):

    if os.environ.get("MAGICS_DUMP_YAML"):
        import yaml

        actions = []
        for n in args:
            if isinstance(n, (list, tuple)):
                for v in n:
                    actions.append(v.to_yaml())
            else:
                actions.append(n.to_yaml())
        print(yaml.dump(dict(plot=actions), default_flow_style=False))
        return

    context.set()
    # try :
    Magics.init()
    for n in args:
        _execute(n)

    # Collect the drivers!
    Magics.finalize()
    for f in context.tmp:
        if os.path.exists(f):
            os.remove(f)
    # except:
    # print ("Magics Error")


def tofortran(file, *args):
    return
    f = open(file + ".f90", "w")
    print(f, "\tprogram magics\n")
    print(f, "\tcall popen\n")
    for n in args:
        n.tofortran(f)
    print(f, "\tcall pclose\n")
    print(f, "\tend")


def tohtml(file, *args):
    return
    f = open(file + ".html", "w")
    print(f, "<html>")
    for n in args:
        n.tohtml(f)
    print(f, "</html>")


def tomv4(file, *args):
    return
    f = open(file + ".mv4", "w")
    for n in args:
        n.tomv4(f)


class odb_filter(object):
    def __init__(self, _m=None, **kw):
        args = {}
        self.verb = "odbfilter"
        if _m is not None:
            args.update(_m)
        self.args = args

    def execute(self, key):
        file = "data%d" % numpy.random.randint(1, 1000)
        odb = "%s.odb" % file
        context.tmp.append(odb)
        cmd = (
            'odbsql -q "'
            + self.args["query"]
            + '" -i '
            + self.args["path"]
            + " -f newodb -o "
            + odb
        )
        print(cmd)
        if os.system(cmd):
            print("Error in filtering ODB data... Aborting")
            os.abort()
        Magics.setc("odb_filename", odb)

    def inspect(self):
        cmd = (
            'odbsql -q "'
            + self.args["query"]
            + '" -i '
            + self.args["path"]
            + " -o data.ascii"
        )
        if os.system(cmd):
            print("Error in filtering ODB data... Aborting")
            os.abort()
        cmd = os.environ["ODB_REPORTER"] + " %s" % "data.ascii"
        if os.system(cmd):
            print("Error in viewing ODB data... Aborting")
            os.abort()


def _jplot(*args):
    from IPython.display import Image

    f, tmp = tempfile.mkstemp(".png")
    os.close(f)

    base, ext = os.path.splitext(tmp)

    img = output(
        output_formats=["png"],
        output_name_first_page_number="off",
        output_name=base,
    )

    all = [img]
    all.extend(args)

    _plot(all)

    image = Image(tmp)
    os.unlink(tmp)
    return image


def plot(*args, **kwargs):
    with LOCK:
        if ipython_active:
            return _jplot(*args, **kwargs)
        else:
            return _plot(*args, **kwargs)


def wmsstyles(data):
    with LOCK:
        context.set()
        try:
            Magics.init()
            styles = data.style()
            Magics.finalize()
            styles = json.loads(styles.decode())
            return styles
        except Exception:
            Magics.finalize()
            return {}


def known_drivers():
    with LOCK:
        return Magics.known_drivers()


def version():
    with LOCK:
        version = Magics.version()
        return version.decode()


def predefined_areas():
    with LOCK:
        home = Magics.home()
        with open("%s/share/magics/styles/projections.json" % (home.decode())) as input:
            projections = json.load(input)
        return list(projections.keys())


def wmscrs():
    os.environ["MAGPLUS_QUIET"] = "on"
    os.environ["MAGPLUS_WARNING"] = "off"
    return {
        "crss": [
            {
                "name": "EPSG:4326",
                "w_lon": -180.0,
                "s_lat": -90.0,
                "e_lon": 180.0,
                "n_lat": 90.0,
            },
            {
                "name": "EPSG:3857",
                "w_lon": -20026376.39,
                "s_lat": -20048966.10,
                "e_lon": 20026376.39,
                "n_lat": 20048966.10,
            },
            {
                "name": "EPSG:32661",
                "w_lon": 1994055.62,
                "s_lat": 5405875.53,
                "e_lon": 2000969.46,
                "n_lat": 2555456.55,
            },
            {
                "name": "EPSG:32762",
                "w_lon": 1999030.54,
                "s_lat": 1444543.45,
                "e_lon": 2005944.38,
                "n_lat": -1405875.53,
            },
            {  # 1896628.62,1507846.05,4662111.45,6829874.45
                # 1896628.62,1507846.05,4662111.45,6829874.45
                "name": "EPSG:3035",
                "w_lon": 1896628.62,
                "s_lat": 1507846.05,
                "e_lon": 4662111.45,
                "n_lat": 6829874.45,
            },
        ],
        "geographic_bounding_box": {
            "w_lon": -180.0,
            "e_lon": 180.0,
            "s_lat": -90.0,
            "n_lat": 90.0,
        },
    }


def get_library_path():
    return Magics.get_library_path()



# (C) Copyright 1996-2016 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.

import os
import tempfile

import simplejson

magics = {}
magics["info"] = "on"

magics["format"] = "a4"
magics["page"] = []
magics["orientation"] = "portrait"
magics["version"] = "3.0"
last = {}

ps = {}
ps["format"] = "ps"

drivers = [ps]
magics["drivers"] = drivers


definition = []
definition.append(
    {
        "id": "haxis",
        "class": "horizontal_axis",
        "axis_type": "date",
        "axis_tick_label": "off",
        "axis_tick_colour": "kelly_green",
        "axis_line_colour": "kelly_green",
        "axis_grid": "on",
        "axis_grid_line_style": "dash",
        "axis_grid_colour": "kelly_green",
    }
)

definition.append(
    {
        "class": "horizontal_axis",
        "id": "haxis_last",
        "axis_type": "date",
        "axis_tick_label": "on",
        "axis_tick_colour": "kelly_green",
        "axis_line_colour": "kelly_green",
        "axis_grid": "on",
        "axis_grid_line_style": "dash",
        "axis_grid_colour": "kelly_green",
        "axis_date_type": "days",
        "axis_minor_tick": "on",
        "axis_days_label": "both",
        "axis_days_label_colour": "navy",
        "axis_days_label_height": "0.35",
        "axis_months_label_colour": "navy",
        "axis_months_label_height": "0.3",
        "axis_years_label_colour": "navy",
        "axis_years_label_height": "0.3",
    }
)


definition.append(
    {
        "class": "vertical_axis",
        "id": "vaxis",
        "axis_line": "on",
        "axis_grid": "on",
        "axis_line_colour": "kelly_green",
        "axis_tick_label_colour": "kelly_green",
        "axis_label_font": "sansserif",
        "axis_grid_line_style": "dash",
        "axis_grid_colour": "kelly_green",
        "axis_tick_colour": "kelly_green",
    }
)
magics["definition"] = definition
pages = []

header = {"height": "12%"}


def page():
    page = {
        "height": "15%",
        "map": {
            "cartesian": {
                "x_date": {"automatic": "on"},
                "y_regular": {"automatic": "on"},
            },
            "text": {
                "left": "0%",
                "height": "10%",
                "bottom": "90%",
                "text_justification": "left",
                "text_font_size": "0.4",
                "text_font_style": "bold",
                "text_font": "sansserif",
                "text_colour": "navy",
                "text_border": "off",
            },
            "horizontal_axis": {"use_id": "haxis"},
            "vertical_axis": {"use_id": "vaxis"},
        },
    }
    return page


class humidity(object):
    def execute(self):
        humi = {
            "epsbufr": {
                "epsbufr_parameter_descriptor": "13003",
                "epsbufr_parameter_title": "850 hPa Relative Humidity  (%)",
                "use_id": "station",
            },
            "metgraph": {"curve": {}},
        }

        map = page()
        humi["epsbufr"]["epsbufr_information"] = magics["info"]
        magics["info"] = "off"
        map["map"]["cartesian"] = {
            "x_date": {"automatic": "on"},
            "y_regular": {"y_min": "0", "y_max": "100"},
        }
        map["map"]["plot"] = humi
        magics["page"].append(map)
        return map


class msl(object):
    def execute(self):
        msl = {
            "epsbufr": {
                "epsbufr_parameter_descriptor": "10051",
                "epsbufr_parameter_title": "MSL Pressure (hPa)",
                "epsbufr_parameter_scaling_factor": "0.01",
                "use_id": "station",
            },
            "metgraph": {"curve": {}},
        }

        map = page()
        msl["epsbufr"]["epsbufr_information"] = magics["info"]
        magics["info"] = "off"
        map["map"]["plot"] = msl
        magics["page"].append(map)
        return map


class cloud(object):
    def execute(self):
        cloud = {
            "epsbufr": {
                "epsbufr_parameter_descriptor": "20010",
                "epsbufr_parameter_title": "Cloud Amount (%)",
                "use_id": "station",
            },
            "metgraph": {"bar": {}},
        }

        map = page()
        cloud["epsbufr"]["epsbufr_information"] = magics["info"]
        magics["info"] = "off"
        map["map"]["cartesian"] = {
            "x_date": {"automatic": "on"},
            "y_regular": {"y_min": "0", "y_max": "100"},
        }
        map["map"]["plot"] = cloud
        magics["page"].append(map)
        return map


class precip(object):
    def execute(self):
        precip = {
            "epsbufr": {
                "epsbufr_parameter_descriptor": "13011",
                "epsbufr_parameter_title": "Precipitation",
                "epsbufr_accumulated_parameter": "on",
                "use_id": "station",
            },
            "metgraph": {"bar": {}},
        }
        precip["epsbufr"]["epsbufr_information"] = magics["info"]
        magics["info"] = "off"
        map = page()
        map["map"]["plot"] = precip
        magics["page"].append(map)
        return map


class wind(object):
    def execute(self):
        wind = {
            "epsbufr": {
                "epsbufr_parameter_descriptor": "11003",
                "epsbufr_parameter_2_descriptor": "11004",
                "epsbufr_parameter_title": "10m Wind (m/s)",
                "use_id": "station",
            },
            "metgraph": {"flags": {}},
        }
        wind["epsbufr"]["epsbufr_information"] = magics["info"]
        magics["info"] = "off"
        map = page()
        map["map"]["plot"] = wind
        map["height"] = "10%"
        map["map"]["height"] = "80%"
        map["map"]["text"]["bottom"] = "70%"
        map["map"]["vertical_axis"]["axis_tick_label"] = "off"
        map["map"]["vertical_axis"]["axis_grid"] = "off"
        map["map"]["vertical_axis"]["axis_tick"] = "off"
        map["map"]["cartesian"] = {
            "x_date": {"automatic": "on"},
            "y_regular": {"y_min": "-1", "y_max": "1"},
        }
        magics["page"].append(map)
        return map


class tempe(object):
    def execute(self):
        t850 = {
            "epsbufr": {
                "epsbufr_parameter_descriptor": "12001",
                "epsbufr_parameter_title": "Temperature (C)",
                "epsbufr_parameter_offset_factor": "-273.15",
                "use_id": "station",
            },
            "metgraph": {"metgram_plot_style": "curve", "metgram_curve_colour": "blue"},
        }
        t850["epsbufr"]["epsbufr_information"] = magics["info"]
        tempe = {
            "epsbufr": {
                "epsbufr_parameter_descriptor": "12004",
                "epsbufr_short_title": "off",
                "epsbufr_information": "off",
                "epsbufr_parameter_offset_factor": "-273.15",
                "use_id": "station",
            },
            "metgraph": {"metgram_plot_style": "curve", "metgram_curve_colour": "red"},
        }
        t850["epsbufr"]["epsbufr_information"] = magics["info"]
        magics["info"] = "off"
        map = page()
        map["map"]["plot"] = [t850, tempe]
        magics["page"].append(map)
        return map


class station(object):
    def __init__(self, args):
        self.definition = args
        self.definition["id"] = "station"
        self.definition["class"] = "epsbufr"

    def execute(self):
        magics["definition"].append(self.definition)
        return page()


class ps(object):
    def __init__(self, args):
        self.definition = args
        self.definition["format"] = "ps"

    def execute(self):
        if magics["drivers"] is None:
            magics["drivers"] = []
        magics["drivers"].append(self.definition)
        return page()


def metgram(*args):

    magics["page"].append(header)
    nb = len(args)
    i = 0
    haxis = "haxis"
    for n in args:
        map = n.execute()
        i += 1
        if i == nb:
            haxis = "haxis_last"
        map["map"]["horizontal_axis"]["use_id"] = haxis

    s = simplejson.dumps(magics, indent=4 * " ")
    f = tempfile.NamedTemporaryFile()
    f.write(s)
    f.flush()

    cmd = "magjson %s" % (f.name)

    error = os.system(cmd)
    if error != 0:
        print("Error found - unix error: ", os.WEXITSTATUS(error))
    f.close


cloud = cloud()
humidity = humidity()
precip = precip()
tempe = tempe()
msl = msl()
wind = wind()



#!/usr/bin/env python
#
# Copyright 2017-2018 European Centre for Medium-Range Weather Forecasts (ECMWF).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import sys

import setuptools


def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return io.open(file_path, encoding="utf-8").read()


version = None
for line in read("Magics/__init__.py").split("\n"):
    if line.startswith("__version__"):
        version = line.split("=")[-1].strip()[1:-1]

assert version

install_requires = ["numpy"]
if sys.version_info < (3, 7):
    install_requires = ["numpy<1.20"]

install_requires += ["findlibs"]

setuptools.setup(
    name="Magics",
    version=version,
    description="Python interface to plot meteorological data in GRIB, NetCDF and BUFR.",
    long_description=read("README.rst"),
    author="European Centre for Medium-Range Weather Forecasts (ECMWF)",
    author_email="software.support@ecmwf.int",
    license="Apache License Version 2.0",
    url="https://github.com/ecmwf/magics-python",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    tests_require=[
        "pytest",
    ],
    test_suite="tests",
    zip_safe=True,
    keywords="magics visualisation meteorology grib netcdf bufr",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Operating System :: OS Independent",
    ],
)
